/** This module provides some reflection capabilities, by referring at run time
 *  to data generated at compile time.
 *  The latter is achieved by using templates to recursively define (const)
 *  struct instances describing types.
 *  
 *  To include vtable and monitor pointers in class bodies (including interface
 *  vtable pointers), compile with version ShowHiddenClassFields enabled.
 *  
 *  Note: This module was $(U only) tested using DMD running on Linux. See below for
 *  some portability concerns.
 *  
 *  Bugs:
 *  The code is currently Tango-only, but that should be relatively easy to
 *  change.
 *  
 *  (for Tango users) This code is affected by
 *  $(LINK2 http://d.puremagic.com/issues/show_bug.cgi?id=887, DMD bug #887).
 *  (Should be fixed in DMD v1.005, AFAIK not yet in Tango)
 *  The workaround is in place, and can be disabled by disabling version 
 *  Bug_887_Workaround which is in place by default. Since the workaround
 *  requires TypeInfo.toString() to be called when it shouldn't be necessary,
 *  RttiData's member functions are probably less efficient than they should be.
 *  
 *  It's also affected by
 *  $(LINK2 http://d.puremagic.com/issues/show_bug.cgi?id=902, DMD bug #902),
 *  which means you can't use it from more than one module until that bug is
 *  fixed.
 *  
 *  Static constructors of alias member parameters are assumed to have been run
 *  before that of the template itself. Fortunately this seems to be exactly
 *  what DMD does.
 *  
 *  Unions are used to avoid some extra static constructors. (Unless disabled
 *  by removing version=UnionHack, but that of course introduces extra static
 *  constructors) These assume that a dynamic array has the exact layout of a
 *  struct { size_t length; void* ptr; }.
 *  
 *  Because of DMD bug #902 you can't instantiate the templates (including
 *  internal ones) from more than one module without getting 'multiple
 *  definition' errors.
 *  
 *  The current version requires Tango, but that can easily be changed.
 *  
 *  Assumes the current DMD layout for dynamic arrays and delegates.
 *  
 *  The extra data generated with version=ShowHiddenClassFields seems to be
 *  correct for DMD, but I have no idea if (a) it is correct in all cases and
 *  (b) if it's correct for GDC at all.
 *  (If anyone can tell me whether this works on GDC please do so.
 *  Specifically, the exact layout of class bodies (including interface vptrs) would
 *  be nice to know)
 *  
 *  Type information does not preserve typedefs.
 *  
 *  Authors: Frits van Bommel
 */

module rtti;

//debug = instantiate;    // trace template instantiations

// Uses a union{array, struct} with an initializer on the struct
// to avoid a static this() initializing the array.
//version = UnionHack;

// Show hidden vptr and monitor pointers (including interface vptrs) for classes
version = ShowHiddenClassFields;

version(Tango) {
    // DMD bug #887: TypeInfo does not correctly override opCmp, toHash
    // This breaks bool[TypeInfo].
    // workaround: use bool[char[]] and use TypeInfo.toString() as key.
    // (Fixed in Phobos, AFAIK not yet in Tango)
    version = Bug_887_Workaround;
}

import tango.core.Tuple;
import tango.text.locale.Format;
import tango.text.convert.Integer;
import tango.util.meta.Traits;
import tango.util.meta.Convert;

import util;

/// Represents a type. At most one of fields and base will be non-null.
struct RttiData {
    /** The TypeInfo of the represented type.
     *  
     *  Note: Class references use the TypeInfo of the class, their bodies are
     *  represented by the TypeInfo of ClassBody!(ClassName).
     */
    const TypeInfo typeinfo;
    
    /** The type of the data pointed to if this instance represents a pointer
     *  or class reference.
     */
    const RttiData* base;
    
    /** The fields of the type.
     *  
     *  If the type is a class body, struct or union, this contains the
     *  _fields of the type.
     *  For dynamic arrays, this contains their length and ptr,
     *  for static arrays their elements, and
     *  for delegates their context pointer and function pointer.
     */
    version (UnionHack) {
        union {
            const FieldData[] fields;
            struct {
                private:
                // hack to allow initializer (doesn't seem to like arrays...)
                const size_t fields_length;
                // ditto
                const FieldData* fields_ptr;
            }
        }
    } else { // !UnionHack
        const FieldData[] fields;
    }
    
    invariant {
        assert(typeinfo !is null);
        assert(!(fields && base));
    }
    
    /** Convert to a pretty-printed string.
     *  
     *  Note: This can get pretty long, especially for types containing 
     *  class references, structs, unions, arrays or pointers.
     *  
     *  Bugs: Not re-entrant, due to the way it avoids infinite recursion.
     */
    char[] toString() {
        // {{{ BEGIN hack to avoid unbounded recursion
        version (Bug_887_Workaround)
            static size_t[char[]] seen;
        else
            static size_t[TypeInfo] seen;
        static size_t depth = 0; // (Note: now also used for pretty-printing)
        
        depth++;
        scope(exit) {
            depth--;
            if (depth == 0) seen = null;
        }
        // }}} END hack to avoid unbounded recursion
        
        version (Bug_887_Workaround)
            char[] key = typeinfo.toUtf8();
        else
            alias typeinfo key;
        
        if (key in seen && fields.length > 0) {
            return "[[" ~ typeinfo.toUtf8() ~ "]]";
        }
        
        seen[key]++;
        scope(exit) {
            seen[key]--;
            if (seen[key] == 0)
                seen.remove(key);
        }
        
        /// Generate spaces for a given indentation level
        static char[] indent(size_t level) {
            static char[] spaces = "    ";
            size_t depth = level * 4;
            while (spaces.length < depth) { spaces ~= spaces; }
            return spaces[0 .. depth];
        }
        
        char[] ret = "RttiData(" ~ typeinfo.toUtf8();
        if (fields.length != 0) {
            ret ~= ", [";
            foreach(idx, f; fields) {
                ret ~= \n ~ indent(depth) ~ f.toString() ~ ", ";
            }
            if (fields.length > 0) ret = ret[0 .. $-2];    // remove last ", "
            ret ~= \n ~ indent(depth - 1) ~ "]";
        } else if (base) {
            ret ~= " --> " ~ base.toString();
        }
        ret ~= ")";
        
        return ret;
    };
    
    /** Returns !=0 if this instance represents the same type as the other 
     *  instance.
     */
    int opEquals(RttiData other) {
        return typeinfo == other.typeinfo;
    }
    
    version (Bug_887_Workaround) {
        ///
        int opCmp(RttiData other) {
            char[] str = typeinfo.toUtf8();
            char[] ostr = other.typeinfo.toUtf8();
            return typeid(char[]).compare(&str, &ostr);
        }
        
        ///
        hash_t toHash() {
            char[] str = typeinfo.toUtf8();
            return typeid(char[]).getHash(&str);
        }
    } else {
        ///
        int opCmp(RttiData other) {
            return typeid(TypeInfo).compare(&typeinfo, &other.typeinfo);
        }
        
        ///
        hash_t toHash() {
            return typeinfo.toHash();
        }
    }
}

/// Represents a field in a class body, struct or union.
struct FieldData {
    /// A pointer to the type data of the type of the field.
    const RttiData* rtti;
    
    /// The _offset of this field from the start of the data block.
    const size_t offset;
    
    invariant {
        assert(rtti);
    }
    
    /** Convert to a pretty-printed string.
     *  
     *  Note: This can get pretty long, especially for fields containing class
     *  references, structs, unions, arrays or pointers.
     *  
     *  Bugs: Not re-entrant because it calls RttiData.toString, which isn't
     *  re-entrant.
     */
    char[] toString() {
        return Formatter.format("@ {0,3}: {1}", 
                        offset, rtti.toString);
    }
    
    ///
    int opEquals(FieldData other) {
        return offset == other.offset && *rtti == *other.rtti;
    }
    
    ///
    int opCmp(FieldData other) {
        if (offset != other.offset)
            return typeid(typeof(offset)).compare(this, &other);
        return rtti.opCmp(other.rtti);
    }
    
    ///
    hash_t toHash() {
        return rtti.toHash() + offset * 31;
    }
}

/// Returns an RttiData instance describing T.
template Rtti(T) {
    debug(instantiate) pragma(msg, "Rtti " ~ T.mangleof);
    
    alias RttiImpl!(T).Rtti Rtti;
//    const Rtti = RttiImpl!(T).Rtti;
}

private template RttiImpl(T) {
    debug(instantiate) pragma(msg, "RttiImpl " ~ T.mangleof);
    
    static if (is(T == struct) || is(T == union)) {
        debug(instantiate) pragma(msg, "RttiImpl : !Class");
        
        alias TupleRtti!(T, FieldsRtti!(T)).Rtti Rtti;
        
    } else static if (is(T == class)) {
        debug(instantiate) pragma(msg, "RttiImpl : Class");
        
        alias TupleRtti!(ClassBody!(T), ClassFieldsRtti!(T)).Rtti BodyRtti;
        
        const RttiData Rtti = {
            typeinfo: typeid(T),
            fields: null,
            base: &BodyRtti
        };
        
    } else static if (is(T Base == Base*)) {
        debug(instantiate) pragma(msg, "RttiImpl : " ~ Base.mangleof ~ "*");
        
        const RttiData Rtti = {
            typeinfo: typeid(T),
            fields: null,
            base: &.Rtti!(Base)
        };
        
    } else static if (is(T Base == Base[])) {
        debug(instantiate) pragma(msg, "RttiImpl : " ~ Base.mangleof ~ "[]");
        
        struct Layout {
            size_t length;
            Base* ptr;
        }
        
        static assert(Layout.sizeof == T.sizeof);
        
        alias TupleRtti!(T, FieldsRtti!(Layout)).Rtti Rtti;
        
    } else static if (is(T Base : Base[])) {
        debug(instantiate) pragma(msg, "RttiImpl : " ~ Base.mangleof ~ "["
                                        ~ uintToString!(T.sizeof/Base.sizeof) ~ "]");
        
        static assert(isStaticArrayType!(T));
        static assert(is(T == Base[T.sizeof/Base.sizeof]));
        static assert(Base.sizeof % Base.alignof == 0);
        
        template Fields(size_t Times, T, size_t offset = 0u) {
            static if (Times == 0) {
                alias Tuple!() Fields;
            } else {
                alias Tuple!(
                    FieldRtti!(T, offset),
                    Fields!(Times - 1, T, offset + T.sizeof)
                ) Fields;
            }
        }
        
        alias TupleRtti!(T, Fields!(T.sizeof / Base.sizeof, Base)).Rtti Rtti;
        
    } else static if (is(T FuncType == delegate)) {
        debug(instantiate) pragma(msg, "RttiImpl : delegate");
        
        static if (is(T Result == return) && is(FuncType Pars == function)) {
            
            struct Layout {
                void* ptr;
                Result function(Pars) funcptr;
            }
            
            static assert(Layout.sizeof == T.sizeof);
            
            alias TupleRtti!(T, FieldsRtti!(Layout)).Rtti Rtti;
            
        } else {
            static assert(0, "Error instantiating Rtti!(delegate): "
                ~ "is-expressions failed? This shouldn't happen!");
        }
        
    } else static if (is(T Base == typedef)) {
        debug(instantiate) pragma(msg, "RttiImpl : typedef " ~ T.mangleof ~ " >> " ~ Base.mangleof);
        
        static assert(is(T == typedef));
        static assert(!is(T == Base));
        
        alias .Rtti!(Base) Rtti;
//        const RttiData Rtti = {
//            typeinfo: typeid(T),
//            fields: .Rtti!(Base).fields,
//            base: .Rtti!(Base).base
//        };
        
    } else static if (isAssociativeArray!(T)) {
        // ICE if placed here:
        debug(instantiate) pragma(msg, "RttiImpl : AA : " ~ ValueType.mangleof ~ "[" ~ KeyType.mangleof ~ "]");
        alias typeof(T.keys[0]) KeyType;
        alias typeof(T.values[0]) ValueType;
        
        debug(instantiate) pragma(msg, "RttiImpl : AA : " ~ ValueType.mangleof ~ "[" ~ KeyType.mangleof ~ "]");
        
        static assert(is(T == ValueType[KeyType]));
        static assert(T.sizeof == (void*).sizeof);
        
        const RttiData Rtti = {
            typeinfo: typeid(T),
            fields: null,
            base: &.Rtti!(void)
        };
        
    } else {
        debug(instantiate) pragma(msg, "RttiImpl : !Aggregate");
        
        const RttiData Rtti = { typeinfo: typeid(T), fields: null };
        
    }
}


/** RttiData instances for class bodies use the TypeInfo of ClassBody,
 *  templated on the class, as their type.
 *  (the class' own TypeInfo is used for the reference)
 *  
 *  For instance, Rtti!(Object).base.typeinfo == typeid(ClassBody!(Object)).
 *  
 *  This type does not have any members, and its only value is the fact that it
 *  has TypeInfo data associated with it.
 */
struct ClassBody(T) {
    static assert(is(T == class));
}

/** Returns a tuple of FieldData structs that represent a class' fields
 *  (including those of base classes) and, with version=ShowHiddenClassFields
 *  enabled, its hidden vptr and monitor pointers.
 */
template ClassFieldsRtti(T) {
    alias ClassFieldsRttiImpl!(T).fields ClassFieldsRtti;
}

private template ClassFieldsRttiImpl(T) {
    debug(instantiate) pragma(msg, "ClassMembers " ~ T.mangleof);
    
    static assert(is(T == class));
    
    static if (is(T == Object)) {
        // End recursion at Object, the base of the class hierarchy
        
        static assert(T.tupleof.length == 0);
        static assert(BaseTypeTuple!(T).length == 0);
        
        version (ShowHiddenClassFields) {
            alias Tuple!(
                FieldRtti!(void*, 0u),  // vptr
                FieldRtti!(void*, (void*).sizeof)   // monitor
            ) fields;
        } else { // !ShowHiddenClassFields
            alias Tuple!() fields;
        }
        
        const size_t size = 2 * (void*).sizeof;
        
    } else {
        alias BaseTypeTuple!(T) Bases;
        
        version (ShowHiddenClassFields) {
            // Add any interface vtprs
            
            // prevsize: the size of the instance before interface vptrs.
            static if (T.tupleof.length > 0) {
                const prevsize = T.tupleof[$-1].offsetof + T.tupleof[$-1].sizeof;
            } else {
                const prevsize = ClassFieldsRttiImpl!(Bases[0]).size;
            }
            
            // Interfaces: Tuple of interface vptr FieldData instances.
            // size: the size of the object body, including interface vptrs.
            //       (needed by derived classes without members to calculate
            //       prevsize)
            static if (Bases.length == 1) {
                // No interfaces implemented.
                // This needs to be handled separately since no alignup occurs.
                alias Tuple!() Interfaces;
                const size = prevsize;
            } else {
                // T implements some interfaces, add interface vptrs
                const if_offset = alignUp!((void*).alignof, prevsize);
                alias InterfacesRtti!(if_offset, Bases[1 .. $]) Interfaces;
                const size = if_offset + (void*).sizeof * (Bases.length - 1);
            }
        } else { // !ShowHiddenClassFields
            alias Tuple!() Interfaces;
        }
        
        // Get base class and recurse, appending fields of current class.
        alias Tuple!(
            ClassFieldsRttiImpl!(Bases[0]).fields,
            FieldsRtti!(T),
            Interfaces
        ) fields;
    }
}

// Given an offset and tuple of interface types, generate FieldData tuple for
// interface vtable pointers
private template InterfacesRtti(size_t offset, Ts...) {
    static assert(offset % (void*).sizeof == 0);
    
    static if (Ts.length == 0) {
        alias Tuple!() InterfacesRtti;
    } else {
        static assert(is(Ts[0] == interface));
        
        alias Tuple!(
            FieldRtti!(void*, offset),
            InterfacesRtti!(offset + (void*).sizeof, Ts[1..$])
        ) InterfacesRtti;
    }
}

/** Returns a FieldData tuple representing the fields of a struct or union.
 *  ClassFieldsRtti should be used to generate such tuples for classes.
 *  (Internally ClassFieldsRtti uses FieldsRtti)
 */
template FieldsRtti(T) {
    debug(instantiate) pragma(msg, "FieldsRtti " ~ T.mangleof 
        ~ " ( " ~ TypeTupleToString!(typeof(T.tupleof)) ~ ")");
    
    alias FieldsRttiImpl!(T) FieldsRtti;
}

private template FieldsRttiImpl(T, size_t Idx = 0u) {
    debug(instantiate) pragma(msg, "FieldsRttiImpl " ~ T.mangleof
        ~ " ( " ~ TypeTupleToString!(typeof(T.tupleof)) ~ ")"
        ~ " [" ~ uintToString!(Idx) ~ "]");
    
    static assert(is(T == class) || is(T == struct) || is(T == union));
    
    static if (Idx < typeof(T.tupleof).length)
        alias Tuple!(
            FieldRtti!(typeof(T.tupleof)[Idx], T.tupleof[Idx].offsetof),
            FieldsRttiImpl!(T, Idx + 1)
        ) FieldsRttiImpl;
    else
        alias Tuple!() FieldsRttiImpl;
}

/** Returns a FieldData instance for a field of type T at offset Offset from the
 *  start of an instance.
 */
template FieldRtti(T, size_t Offset) {
    debug(instantiate) 
        pragma(msg, "FieldRtti " ~ T.mangleof ~ " @ " ~ uintToString!(Offset));
    alias FieldRttiImpl!(T, Offset).FieldRtti FieldRtti;
}

private template FieldRttiImpl(T, size_t Offset) {
    const FieldData FieldRtti = {
//        rtti: &Rtti!(T),      // recursive alias declaration ? :(
        offset: Offset//,
    };
    static this() { FieldRtti.rtti = &Rtti!(T); }
}

// Used to implement Rtti for structs, dynamic and static arrays, delegates.
private template TupleRtti(T, Fields...) {
    private alias Array!(FieldData, Fields) Fs;
    
    version (UnionHack) {
        debug(instantiate) pragma(msg, "TupleRtti : UnionHack");
        const RttiData Rtti = {
            typeinfo: typeid(T),
            fields_length: Fs.length,
            fields_ptr: Fs.ptr
        };
    } else { // !UnionHack
        debug(instantiate) pragma(msg, "TupleRtti : !UnionHack");
        
        const RttiData Rtti = { typeinfo: typeid(T) };
        
        static this() {
            Rtti.fields = Fs.ptr[0 .. Fs.length];
        }
    }
}


version (UnitTest) {
    import tango.io.Stdout;
    
    template TD(T) { typedef T TD; }
    template Alias(T) { alias T Alias; }
    
    class Class {}
    struct Struct{}
    union Union {}
    enum Enum { dummy }
    typedef int Typedef;
    alias Tuple!(bool, int, char, long, int[int], Class, Struct, Union, Enum, Typedef) NonVoidTypes;
    alias Tuple!(void, NonVoidTypes) Types;
    
    //
    // Derived Data Types
    //
    
    unittest {
        //    * pointer
        foreach (T; Types) {
            assert(Rtti!(T*).typeinfo == typeid(T*));
            assert(Rtti!(T*).base !is null);
            assert(Rtti!(T*).base == &Rtti!(T));
            assert(Rtti!(T*).fields.length == 0);
        }
    }
    
    unittest {
        //    * array
        foreach (T; Types) {
            
            assert(&Rtti!(T[]));
            assert(Rtti!(T[]).typeinfo == typeid(T[]));
            assert(Rtti!(T[]).base is null);
            assert(Rtti!(T[]).fields.length == 2);
            assert(Rtti!(T[]).fields[0].rtti !is null);
            assert(*Rtti!(T[]).fields[0].rtti == Rtti!(typeof((T[]).length)));
            assert(Rtti!(T[]).fields[0].offset == 0u);
            assert(*Rtti!(T[]).fields[1].rtti == Rtti!(typeof((T[]).ptr)));
            assert(Rtti!(T[]).fields[1].offset == size_t.sizeof);
        }
    }
    
    unittest {
        //    * associative array
        foreach (K; NonVoidTypes[1 .. $]) { // no bool keys
            foreach (V; NonVoidTypes) {
                alias V[K] T;
                static assert(isAssociativeArray!(T));
                assert(Rtti!(T).typeinfo == typeid(T));
                assert(Rtti!(T).base !is null);
                assert(Rtti!(T).base == &Rtti!(void));
                assert(Rtti!(T).fields == null);
            }
        }
    }
    
    unittest {
        //    * function
        
        foreach (R; Types) {
            alias R function() T;
            
            assert(Rtti!(T).typeinfo == typeid(T));
            assert(Rtti!(T).base !is null);
            assert(Rtti!(T).base == &Rtti!(typeof(*T)));
            assert(Rtti!(T).fields == null);
        }
    }
    
    unittest {
        //    * delegate 
        
        foreach (R; Types) {
            alias R delegate() T;
            T t;
            assert(Rtti!(T).typeinfo == typeid(T));
            assert(Rtti!(T).base is null);
            assert(Rtti!(T).fields[0].rtti == &Rtti!(typeof(t.ptr)));
            assert(Rtti!(T).fields[0].offset == 0u);
            assert(Rtti!(T).fields[1].rtti == &Rtti!(typeof(t.funcptr)));
            assert(Rtti!(T).fields[1].offset == (void*).sizeof);
        }
    }
    
    //
    // User Defined Types:
    //
    
    unittest {
        //    * alias
        
        foreach (T; Types) {
            static if (is(T == Typedef)) {
                // Special case: T is a typedef.
                assert(Rtti!(Alias!(T)).typeinfo == typeid(T)
                    || Rtti!(Alias!(T)).typeinfo == typeid(int));
            } else {
                assert(Rtti!(Alias!(T)).typeinfo == typeid(T));
            }
            assert(Rtti!(Alias!(T)).fields == Rtti!(T).fields);
        }
    }
    
    unittest {
        //    * typedef
        
        foreach (T; Types) {
            static if (!is(T == Typedef)) {
                assert(Rtti!(TD!(T)).typeinfo == typeid(TD!(T))
                    || Rtti!(TD!(T)).typeinfo == typeid(T));
            } else {
                // Special case: T is itself a typedef.
                assert(Rtti!(TD!(T)).typeinfo == typeid(TD!(T))
                    || Rtti!(TD!(T)).typeinfo == typeid(int));
            }
            assert(Rtti!(TD!(T)).fields == Rtti!(T).fields);
        }
    }
    
    unittest {
        //    * enum
        
        assert(Rtti!(Enum).typeinfo == typeid(Enum));
        assert(Rtti!(Enum).base == null);
        assert(Rtti!(Enum).fields.length == 0);
    }
    
    unittest {
        //    * struct
        
        assert(Rtti!(Struct).typeinfo == typeid(Struct));
        assert(Rtti!(Struct).base == null);
        assert(Rtti!(Struct).fields.length == 0);
        
        struct TestStruct {
            int i;
            double d;
            byte b;
            int i2;
            short s;
            byte b2;
            long l;
        }
        
        assert(Rtti!(TestStruct).typeinfo == typeid(TestStruct));
        assert(Rtti!(TestStruct).base == null);
        assert(Rtti!(TestStruct).fields.length == TestStruct.tupleof.length);
        foreach (i, T; typeof(TestStruct.tupleof)) {
            assert(*Rtti!(TestStruct).fields[i].rtti == Rtti!(T));
            assert(Rtti!(TestStruct).fields[i].offset == TestStruct.tupleof[i].offsetof);
        }
    }
    
    unittest {
        //    * union
        
        assert(Rtti!(Union).typeinfo == typeid(Union));
        assert(Rtti!(Union).base == null);
        assert(Rtti!(Union).fields.length == 0);
        
        union TestUnion {
            int i;
            double d;
            byte b;
            int i2;
            short s;
            byte b2;
            long l;
        }
        
        assert(Rtti!(TestUnion).typeinfo == typeid(TestUnion));
        assert(Rtti!(TestUnion).base == null);
        assert(Rtti!(TestUnion).fields.length == TestUnion.tupleof.length);
        foreach (i, T; typeof(TestUnion.tupleof)) {
            assert(*Rtti!(TestUnion).fields[i].rtti == Rtti!(T));
            assert(Rtti!(TestUnion).fields[i].offset == TestUnion.tupleof[i].offsetof);
        }
    }
    
    unittest {
        //    * class
        
        version(ShowHiddenClassFields)
            static const size_t hidden = 2;
        else
            static const size_t hidden = 0;
        
        assert(Rtti!(Class).typeinfo == typeid(Class));
        assert(Rtti!(Class).base != null);
        assert(Rtti!(Class).base.typeinfo == typeid(ClassBody!(Class)));
        assert(Rtti!(Class).base.base == null);
        assert(Rtti!(Class).base.fields.length == hidden);
        assert(Rtti!(Class).fields.length == 0);
        
        static class TestClass {
            int i;
            double d;
            byte b;
            int i2;
            short s;
            ubyte b2;
            long l;
        }
        
        assert(Rtti!(TestClass).typeinfo == typeid(TestClass));
        assert(Rtti!(TestClass).base != null);
        assert(Rtti!(TestClass).fields.length == 0);
        RttiData* cbody = Rtti!(TestClass).base;
        assert(cbody.typeinfo == typeid(ClassBody!(TestClass)));
        assert(cbody.base == null);
        assert(cbody.fields.length == TestClass.tupleof.length + hidden);
        foreach (i, T; typeof(TestClass.tupleof)) {
            assert(*cbody.fields[i+hidden].rtti == Rtti!(T));
            assert(cbody.fields[i+hidden].offset == TestClass.tupleof[i].offsetof);
        }
    }
}
